/* Fold.e, by Larry Margolis

  Code for folding of procedures.  Place the cursor on a brace or parenthesis,
  and enter FOLD in the EPM command dialog.  The code between the character
  the cursor is on and the matching bracket will be folded (actually, displayed
  in a 1-point type).  Place the cursor inside the folded text and enter the
  EXPAND command to restore it.

  Installation:  This can be included in your base .ex files by adding the line:
     include 'fold.e'
  to your MYSTUFF.E or MYEXTRA.E, or compiled as a separate .ex file and linked
  when you want to use it.  If compiled separately, you can also enter the FOLD
  command to fold text without linking fold.ex, and FOLD 0 or FOLD OFF to
  expand folded text.  (That's what the DEFMAIN accomplishes.)
*/

compile if EVERSION < '5.50'
   Error:  This only supports EPM 5.50 or above - mixed font support is required.
compile endif

const GOLD = '(){}[]<>'  -- Parens, braces, brackets & angle brackets.

compile if not defined('SMALL')  -- being compiled as a stand-alone module
defmain
   uparg = upcase(arg(1))
   if uparg=0 | uparg='OFF' then
      'expand'
   elseif uparg=1 | uparg='ON' | uparg='' then
      'fold'
   else
      gold2='     '
      do i=1 to length(GOLD)
         gold2 = gold2 substr(GOLD, i, 1)
      enddo
      call winmessagebox('Fold', 'Place cursor on one of:'\n gold2\n'then enter FOLD to fold text.  Place cursor inside folded text and enter FOLD OFF to expand it.', 16384+48) -- MOVEABLE + MB_INFORMATION
   endif
compile endif

const
   COLOR_CLASS = 1
   BOOKMARK_CLASS = 13
   STYLE_CLASS =  14
   FONT_CLASS =  16

   FIND_NEXT_ATTR_SUBOP = 1
   FIND_PREV_ATTR_SUBOP = 2
   FIND_MATCH_ATTR_SUBOP = 3
   DELETE_ATTR_SUBOP = 16

defc fold =
   call psave_pos(savepos)
   n=1
   c=substr(textline(.line),.col,1)
   GETSEARCH search_command -- Save user's search command.
   k=pos(c,GOLD)            --  '(){}[]<>'
   if not k then sayerror 'Not a balanceable character.'; return; endif
   fstline=.line; fstcol=.col
   search = substr(GOLD,(k+1)%2*2-1,2)
   if k//2 then direction='+F'; else direction='-R'; endif
   if search='[]' then search='\[\]'; endif
compile if EVERSION >= '5.60'
   if search='()' then search='\(\)'; endif
   'xcom L /['search']/ex'direction     -- Use extended grep; much faster.
compile else
   'xcom L /['search']/eg'direction
compile endif
   loop
      repeatfind
      if rc then leave; endif
      if substr(textline(.line), .col, 1) = c then n=n+1; else n=n-1; endif
      if n=0 then leave; endif
   endloop
   if rc=sayerror('String not found') then
      sayerror 'Unbalanced token.'
   else
      sayerror 1
   endif
   SETSEARCH search_command -- Restores user's command so Ctrl-F works.
   lstline = .line; lstcol = .col
   if not rc then
      getfileid fid
      fontid=registerfont('Courier', 1, 0)
      call attribute_on(4)  -- Mixed fonts flag
      if not (k//2) then  -- adjust the order
         parse value fstline lstline fstcol lstcol with lstline fstline lstcol fstcol
      endif
      fstcol = fstcol + 1        -- Want to display the brace, paren, etc.
      if leftstr(textline(lstline), lstcol-1) = '' then
         lstcol = 0
      else
         lstcol = lstcol - 1
      endif
;getfileid fld, 'fold.e'
;insertline '; k='k'; k//2='k//2'; not1='(not k//2)'; not2='(not (k//2))'; Insert_Attribute_Pair(16,' fontid',' fstline',' lstline',' fstcol',' lstcol',' fid')', fld.last+1, fld
      oldmod = .modify
      Insert_Attribute_Pair(FONT_CLASS, fontid, fstline, lstline, fstcol, lstcol, fid)
      .modify = oldmod
   endif
   call prestore_pos(savepos)

defc expand =
   fontid=registerfont('Courier', 1, 0)
   class = 16  -- STYLE_CLASS
   line=.line; col=.col; offst=0
   do forever
      attribute_action FIND_PREV_ATTR_SUBOP, class, offst, col, line
      if class=0 then   -- not found
         sayerror 'Not found - cursor must be after start of folded text.'
         return
      endif
      query_attribute class, val, IsPush, offst, col, line
      if IsPush & val=fontid then leave; endif
      offst = offst-1
   enddo
              -- OK, at this point we've got the beginning attribute.
   fstline = line; fstcol = col; fstoff = offst
   oldmod = .modify
   attribute_action FIND_MATCH_ATTR_SUBOP, class, offst, col, line
   if class then
      attribute_action DELETE_ATTR_SUBOP, class, offst, col, line
   endif
   class = 16  -- STYLE_CLASS
   attribute_action DELETE_ATTR_SUBOP, class, fstoff, fstcol, fstline
   .modify = oldmod

