/*************************************************************************

                      +---------------------------+
                      |  E P M P R T . E          |
                      +---------------------------+

  Printer Control Code selector for the EPM editor.

  This proc  creates a pulldown menu  of printer  control codes  which you can
  select and imbed in your text  or send  immediately to  the printer.

  You can change the control codes by typing  over them  below, or you
  can add  new ones  by adding  lines using   the template   below.
  The  codes  provided  here  should   work  on  the
  proprinter and graphics printer.

Originally appended to EPM PROCS by 72412757 at EHONE.

Updated by Larry Margolis for use in CUSTEPM.  This is now set up to be
invoked by typing the name of the .ex file on the EPM command line.
The DEFMAIN will update the menu, and then we're done - no code from
this file needs to remain loaded.  (Requires the latest base set of
macros, which includes a KEYIN command defined in STDCTRL.E.)

***************************************************************************/

const
   AF_CHAR        =   1   -- key style constants (from STDCONST.E)
   AF_ALT         =  32
compile if EVERSION < '5.21'
   TILDE_CHAR = ''
compile else
   TILDE_CHAR = '~'
compile endif

defmain
   universal activemenu, defaultmenu
   universal activeaccel
   universal printer_codes_menu_flag
   if printer_codes_menu_flag then
      deletemenu defaultmenu, 9, 0, 0
      call maybe_show_menu()
      printer_codes_menu_flag = 0
      return
   endif
   deletemenu defaultmenu, 6, 0, 0  -- delete the existing Help menu (we want it to stay at the right)
   call add_printer(defaultmenu)
   printer_codes_menu_flag = 1
   call readd_help_menu()
   buildacceltable activeaccel, 'dokey a+P', AF_CHAR+AF_ALT,    80, 998  -- a+P
   buildacceltable activeaccel, 'dokey a+P', AF_CHAR+AF_ALT,   112, 999  -- a+p
   activateacceltable  activeaccel

defproc add_printer(menuname)
   buildsubmenu menuname, 9, TILDE_CHAR'Printer ', '', 0, 0
   buildmenuitem menuname, 9, 901, 'Use Nr. Letter qual.', 'keyin G',    0, 0  -- asc(27,71)
   buildmenuitem menuname, 9, 902, 'End Nr. Letter qual.', 'keyin H',    0, 0  -- asc(27,72)
   buildmenuitem menuname, 9, 903, 'Use 12 CPI Printing' , 'keyin :',    0, 0  -- asc(27,58)
   buildmenuitem menuname, 9, 904, 'End 12 CPI Printing' , 'keyin ',     0, 0  -- asc(18)
   buildmenuitem menuname, 9, 905, 'Use Compressed FONT' , 'keyin ',     0, 0  -- asc(15)
   buildmenuitem menuname, 9, 906, 'End Compressed FONT' , 'keyin ',     0, 0  -- asc(18)
   buildmenuitem menuname, 9, 907, 'Use NORMAL FONT'     , 'keyin I0',   0, 0  -- asc(27,73,48)
   buildmenuitem menuname, 9, 908, 'Emphasized Print'    , 'keyin E',    0, 0  -- asc(27,69)
   buildmenuitem menuname, 9, 909, 'End Emphasized Print', 'keyin F',    0, 0  -- asc(27,70)
   buildmenuitem menuname, 9, 910, 'Form Feed'           , 'keyin ',     0, 0  -- asc(12)
   buildmenuitem menuname, 9, 911, '6 LPI spacing'       , 'keyin A2', 0, 0  -- asc(27,65,12,27,50)
   buildmenuitem menuname, 9, 912, '8 LPI spacing'       , 'keyin 0',    0, 0  -- asc(27,48)
   buildmenuitem menuname, 9, 913, '10 LPI spacing'      , 'keyin 1',    0, 0  -- asc(27,49)
   buildmenuitem menuname, 9, 914, '1-1/2_space'         , 'keyin A2', 0, 0  -- asc(27,65,18,27,50)
   buildmenuitem menuname, 9, 915, 'Double space'        , 'keyin A2', 0, 0  -- asc(27,65,24,27,50)
   buildmenuitem menuname, 9, 916, 'Skip perforation ON' , 'keyin N',    0, 0  -- asc(27,78)
   buildmenuitem menuname, 9, 917, 'Skip perforation OFF', 'keyin O',    0, 0  -- asc(27,79)
   buildmenuitem menuname, 9, 918, 'Stop at end of page' , 'keyin 9',    0, 0  -- asc(27,57)
   buildmenuitem menuname, 9, 919, 'Continue at page end', 'keyin 8',    0, 0  -- asc(27,56)
   buildmenuitem menuname, 9, 920, 'Compressed ON'       , 'keyin ',     0, 0  -- asc(15)
   buildmenuitem menuname, 9, 921, 'Compressed OFF'      , 'keyin ',     0, 0  -- asc(18)
   buildmenuitem menuname, 9, 922, 'Underline ON'        , 'keyin -1',   0, 0  -- asc(27,45,49)
   buildmenuitem menuname, 9, 923, 'Underline OFF'       , 'keyin -0',   0, 0  -- asc(27,45,48)
   buildmenuitem menuname, 9, 924, 'Double Width ON'     , 'keyin W1',   0, 0  -- asc(27,87,49)
   buildmenuitem menuname, 9, 925, 'Double Width OFF'    , 'keyin W0',   0, 0  -- asc(27,87,48)
   buildmenuitem menuname, 9, 926, 'Superscript ON'      , 'keyin S0',   0, 0  -- asc(27,83,48)
   buildmenuitem menuname, 9, 927, 'Subscript ON'        , 'keyin S1',   0, 0  -- asc(27,83,49)
   buildmenuitem menuname, 9, 928, 'Super/subscript OFF' , 'keyin T',    0, 0  -- asc(27,84)

